/*
 * test.c - (c) 1999 Nicolai Haehnle - Locinga@pfaffenwinkel.de
 *
 *   This software is placed in the public domain and can be used freely
 *   for any purpose. It comes without any kind of warranty, either
 *   expressed or implied, including, but not limited to the implied
 *   warranties of merchantability or fitness for a particular purpose.
 *   Use it at your own risk. the author is not responsible for any damage
 *   or consequences raised by use or inability to use this program.
 *
 * This is a demonstration of LibGGIDM's functions and can be used as a 
 * reference programming example.
 *
 * Try scrolling with the arrow keys!
 */

#include <stdio.h>
#include <string.h>

#include <sys/time.h>
#include <unistd.h>

#include <ggi/ggi.h>
#include <ggi/keyboard.h>
#include <ggi/dm.h>


/*
 * Some "game" dependent definitions
 */
#define MAP_WIDTH		10
#define MAP_HEIGHT		10

#define BLOCK_WIDTH		100
#define BLOCK_HEIGHT		100

#define SCREENRES_X		640
#define SCREENRES_Y		480

#define BLOCKS_PER_SCREEN_X	((SCREENRES_X+BLOCK_WIDTH-1) / BLOCK_WIDTH)
#define BLOCKS_PER_SCREEN_Y	((SCREENRES_Y+BLOCK_HEIGHT-1) / BLOCK_HEIGHT)

#define MAX_SCROLLPOS_X		((MAP_WIDTH * BLOCK_WIDTH) - SCREENRES_X)
#define MAX_SCROLLPOS_Y		((MAP_HEIGHT * BLOCK_HEIGHT) - SCREENRES_Y)

/*
 * This is the "map". 0 stands for a black block, 1 for a white block
 */
char Map[MAP_HEIGHT][MAP_WIDTH] = {
	{ 0, 0, 0, 0, 0, 1, 1, 1, 1, 1 },
	{ 0, 1, 1, 1, 0, 1, 0, 0, 0, 1 },
	{ 1, 0, 1, 0, 1, 0, 1, 0, 1, 0 },
	{ 1, 1, 1, 1, 1, 0, 0, 0, 0, 0 },
	{ 1, 1, 0, 1, 1, 0, 0, 1, 0, 0 },
	{ 0, 1, 0, 1, 0, 0, 0, 1, 0, 1 },
	{ 1, 1, 1, 0, 0, 0, 1, 1, 1, 1 },
	{ 0, 1, 0, 0, 0, 1, 0, 0, 1, 1 },
	{ 0, 0, 1, 1, 1, 1, 1, 1, 1, 0 },
	{ 0, 0, 0, 0, 1, 0, 0, 0, 1, 0 } };

/*
 * Three drawmaps:
 *  dmPrimary is the currently visible drawmap.
 *  dmBackbuffer is - the name says it - the back buffer. All drawing is done
 *   in the back buffer. After drawing has finished, the contents are made
 *   visible by ggiDMFlip()
 *  dmBackgrounds contains the "graphics art" used to compose the picture.
 *   The picture for the first block type (id 0) is stored at (0,0),
 *   the picture for the second block type (id 1) is stored at (0,BLOCK_HEIGHT)
 */
ggidm_drawmap dmPrimary, dmBackbuffer, dmBackgrounds;

/*
 * This is the default visual. libGGIDM is attached to it
 */
ggi_visual_t vis;

/*
 * mainLoop() is the actual working function. It is called from main(),
 * after all drawmaps are set up.
 */
void mainLoop()
{
	/*
	 * Is set to 1 when the Escape key is pressed
	 */
	int quit = 0;
	
	/*
	 * How many pixels is the upper-left corner of the screen from the
	 * upper-left corner of the map?
	 */
	int scrollx = 0, scrolly = 0;
	
	/*
	 * Some working variables, used in the drawing loop
	 */
	int firstx, firsty;
	int lastx, lasty;
	int ix, iy;
	
	/*
	 * Variables used to calculate frame length and rate
	 */
	struct timeval lastframe, t;
	int frameno = 0;
	long usecs, average = 0, maxtime = 0;
	unsigned int fphs, minfps = (unsigned int)-1;
	
	/*
	 * Stores which key was pressed
	 */
	int key;
	
	/*
	 * Used to examine the return values of some function calls
	 */
	int error;
	
	/*
	 * Prepare frame calculation
	 */
	gettimeofday(&lastframe, 0);

	do {
		/* 
		 * Check for keypresses
		 */
		while(ggiKbhit(vis)) {
			key = ggiGetc(vis);
			
			switch(key) {
			 case GIIUC_Escape:
			 	/*
			 	 * Set the quit flag. The main loop will
			 	 * terminate when the end is reached
			 	 */
			 	quit = 1;
			 	break;
			 	
			 /*
			  * Scroll the screen if one of the arrow keys
			  * was pressed (also check bounds)
			  */
			 case GIIK_Left:
			 	if (scrollx)
			 		scrollx--;
			 	break;
			 case GIIK_Right:
			 	if (scrollx < MAX_SCROLLPOS_X)
			 		scrollx++;
			 	break;
			 case GIIK_Up:
			 	if (scrolly)
			 		scrolly--;
			 	break;
			 case GIIK_Down:
			 	if (scrolly < MAX_SCROLLPOS_Y)
			 		scrolly++;
			 	break;
			}
		}
		
		/* 
		 * Now for the drawing...
		 *
		 * Transform the scroll position from pixel units into
		 * block units
		 */
		firstx = scrollx / BLOCK_WIDTH;
		firsty = scrolly / BLOCK_HEIGHT;
		lastx = firstx + BLOCKS_PER_SCREEN_X;
		if (lastx >= MAP_WIDTH)
			lastx = MAP_WIDTH - 1;
		lasty = firsty + BLOCKS_PER_SCREEN_Y;
		if (lasty >= MAP_HEIGHT)
			lasty = MAP_HEIGHT - 1;
		
		/*
		 * The drawing loop. It calls ggiDMBlit() for all
		 * blocks that might be visible. ggiDMBlit() automatically
		 * does the needed clipping
		 */
		for(iy = firsty; iy <= lasty; iy++) {
			for(ix = firstx; ix <= lastx; ix++) {
				if ((error = ggiDMBlit(dmBackbuffer, 
				 (ix*BLOCK_WIDTH)-scrollx, 
				 (iy*BLOCK_HEIGHT)-scrolly,
				 dmBackgrounds, 0, BLOCK_HEIGHT*Map[iy][ix],
				 BLOCK_WIDTH, BLOCK_HEIGHT, 0))) {
					ggiPanic("ggiDMBlit failed (%i)\n",
						error);
				}
			}
		}
		
		/* 
		 * Now give the user some information about program speed
		 */
		frameno++;
		gettimeofday(&t, 0);
		usecs = t.tv_usec - lastframe.tv_usec;
		usecs += 1000000 * (t.tv_sec - lastframe.tv_sec);
		if (usecs > maxtime)
			maxtime = usecs;
		average = ((255*average)+usecs) / 256;
		fphs = 100000000 / average;
		if (fphs < minfps)
			minfps = fphs;
		
		memcpy(&lastframe, &t, sizeof(t));
		printf("Frame %i - difftime = %lu, avg=%lu - FPS=%i.%02i\n", 
			frameno, usecs, average, fphs/100, fphs%100);
		
		/* 
		 * Call ggiDMFlip() to make the drawing results visible.
		 * Note that ggiDMFlip() automatically calls ggiFlush() when
		 * in asynchronous mode
		 */
		ggiDMFlip(vis);
	} while(!quit);
	
	/*
	 * Give us some additional information
	 */
	printf("Maximum frame length=%lu - minimum FPS=%i.%02i\n", maxtime,
		minfps/100, minfps%100);
}


int main(void)
{
	/*
	 * We want to use drawing primitives to create the picture of the
	 * blocks. To do so, a visual is needed - ggiDMGetVisual() is called 
	 * to get a sub-visual representing dmBackgrounds.
	 */
	ggi_visual_t vis2;
	
	/*
	 * Used to examine the return value of some function calls
	 */
	int error;
	
	/*
	 * Some variables for palette calculations
	 */
	ggi_color map;
	ggi_pixel white;
	
	
	/* Initialize LibGGI */
	if (ggiInit() != 0) {
		fprintf(stderr, "Unable to initialize LibGGI\n");
		exit(1);
	}
	
	/* Initialize GGI/DM */
	if (ggiDMInit() != 0) {
		ggiPanic("Unable to initialize LibGGIDM\n");
	}
	
	/* Open the default visual */
	if ((vis = ggiOpen(NULL)) == NULL) {
		ggiPanic("Unable to open default visual\n");
	}
	
	/* Attach LibGGIDM */
	if (ggiDMAttach(vis) < 0) {
		ggiPanic("Unable to attach GGIDM\n");
	}
	
	/*
	 * Create dmBackgrounds. Note that it isn't acquired yet, so
	 * drawing operations on it will fail.
	 */
	if (!(dmBackgrounds = ggiDMAddDrawmap(vis, BLOCK_WIDTH,
		               BLOCK_HEIGHT*2, GGIDM_STORE_VIDEOMEMORY))) {
		ggiPanic("Unable to create background drawmap\n");
	}
	
	/* 
	 * Set graphics mode. This also acquires dmBackgrounds (and any
	 * other additional drawmap created before)
	 */
	if (ggiDMSetMode(vis, SCREENRES_X, SCREENRES_Y, 2, GT_8BIT)) {
		ggiPanic("SetMode() failed\n");
	}
	
	/* Set the palette */
	ggiSetColorfulPalette(vis);
	
	/* Get the visible drawmaps */
	if (!(dmPrimary = ggiDMGetVisibleDrawmap(vis, 0))) {
		ggiPanic("GetVisibleDrawmap() failed\n");
	}
	
	if (!(dmBackbuffer = ggiDMGetVisibleDrawmap(vis, 1))) {
		ggiPanic("GetVisibleDrawmap() failed\n");
	}
	
	/*** Draw something in the background ***/
	/*
	 * Get the visual representing dmBackgrounds
	 * No need to set a mode or palette.
	 */
	if (!(vis2 = ggiDMGetVisual(dmBackgrounds))) {
		ggiPanic("Can't get visual for dmBackgrounds\n");
	}

	/* Find color white and make it the foreground color */
	map.r = map.g = map.b = 0xFFFF;
	white = ggiMapColor(vis, &map);
	printf("white=%d\n", white);
	
	ggiSetGCForeground(vis2, white);
	
	/*
	 * Draw something representing block type 1 - a white box
	 */
	ggiDrawBox(vis2, 0, BLOCK_HEIGHT, BLOCK_WIDTH, BLOCK_HEIGHT);

	/* Destroy the visual */
	ggiClose(vis2);
	
	/* Set asynchronous mode in the main visual */
	ggiAddFlags(vis, GGIFLAG_ASYNC);
	
	/* Initalization is done - now for the real thing */
	mainLoop();
	
	/* Detach LibGGIDM */
	ggiDMDetach(vis);
	
	/* Close visual */
	ggiClose(vis);
	
	/* Deinitialize GGI/DM */
	ggiDMExit();
	
	/* Deinitialize LibGGI */
	ggiExit();
	
	return 0;
}
